"""
This module defines some useful design patterns.
"""

from __future__ import division

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__status__ = "Production"
__date__ = "Sep 23, 2011"


class Enum(set):
    """
    Creates an enum out of a set.
    """
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError


class AttrDict(dict):
    """
    Allows to access dict keys as obj.foo in addition to the traditional way
    obj['foo']"
    """
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def copy(self):
        newd = super(AttrDict, self).copy()
        return self.__class__(**newd)


class NullFile(object):
    """A file object that is associated to /dev/null."""
    def __init__(self):
        import os
        return open(os.devnull, 'w')


class NullStream(object):
    """A fake stream with a no-op write.."""
    def write(*args):
        """no-op"""


class FrozenDict(dict):
    """A dictionary that does not permit to redefine its keys."""
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __setitem__(self, key, val):
        if key in self:
            raise KeyError("Cannot overwrite existent key: %s" % str(key))

        dict.__setitem__(self, key, val)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v



"""Tools and helper functions for abinit calculations"""
from __future__ import print_function, division

import os
import collections
import shutil
import operator

from pymatgen.util.string_utils import list_strings, StringColorizer, WildCard

import logging
logger = logging.getLogger(__name__)


class File(object):
    """
    Very simple class used to store file basenames, absolute paths and directory names.
    Provides wrappers for the most commonly used functions defined in os.path.
    """
    def __init__(self, path):
        self._path = os.path.abspath(path)

    def __repr__(self):
        return "<%s at %s, %s>" % (self.__class__.__name__, id(self), self.path)

    def __str__(self):
        return "<%s, %s>" % (self.__class__.__name__, self.path)

    def __eq__(self, other):
        if other is None: return False
        self.path == other.path
                                       
    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def path(self):
        """Absolute path of the file."""
        return self._path

    @property
    def basename(self):
        """File basename."""
        return os.path.basename(self.path)

    @property
    def relpath(self):
        """Relative path."""
        return os.path.relpath(self.path)

    @property
    def dirname(self):
        """Absolute path of the directory where the file is located."""
        return os.path.dirname(self.path)

    @property
    def exists(self):
        "True if file exists."
        return os.path.exists(self.path)

    @property
    def isncfile(self):
        "True if self is a NetCDF file"
        return self.basename.endswith(".nc")

    def read(self):
        """Read data from file."""
        with open(self.path, "r") as f:
            return f.read()

    def readlines(self):
        """Read lines from files."""
        with open(self.path, "r") as f:
            return f.readlines()

    def write(self, string):
        """Write string to file."""
        self.make_dir()
        with open(self.path, "w") as f:
            return f.write(string)

    def writelines(self, lines):
        """Write a list of strings to file."""
        self.make_dir()
        with open(self.path, "w") as f:
            return f.writelines()

    def make_dir(self):
        """Make the directory where the file is located."""
        if not os.path.exists(self.dirname):
            os.makedirs(self.dirname)

    def remove(self):
        """Remove the file."""
        try:
            os.remove(self.path)
        except:
            pass


class Directory(object):
    """
    Very simple class that provides helper functions
    wrapping the most commonly used functions defined in os.path.
    """
    def __init__(self, path):
        self._path = os.path.abspath(path)

    def __repr__(self):
        return "<%s at %s, %s>" % (self.__class__.__name__, id(self), self.path)

    def __str__(self):
        return "<%s, %s>" % (self.__class__.__name__, self.path)

    def __eq__(self, other):
        if other is None: return False
        self.path == other.path

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def path(self):
        """Absolute path of the directory."""
        return self._path

    @property
    def relpath(self):
        """Relative path."""
        return os.path.relpath(self.path)

    @property
    def basename(self):
        """Directory basename."""
        return os.path.basename(self.path)

    def path_join(self, *p):
        """
        Join two or more pathname components, inserting '/' as needed.
        If any component is an absolute path, all previous path components will be discarded.
        """
        return os.path.join(self.path, *p)

    @property
    def exists(self):
        """True if file exists."""
        return os.path.exists(self.path)

    def makedirs(self):
        """
        Super-mkdir; create a leaf directory and all intermediate ones.
        Works like mkdir, except that any intermediate path segment (not
        just the rightmost) will be created if it does not exist.
        """
        if not self.exists:
            os.makedirs(self.path)

    def rmtree(self):
        """Recursively delete the directory tree"""
        shutil.rmtree(self.path, ignore_errors=True)

    def path_in(self, filename):
        """Return the absolute path of filename in the directory."""
        return os.path.join(self.path, filename)

    def list_filepaths(self, wildcard=None):
        """
        Return the list of absolute filepaths in the directory.

        Args:
            wildcard:
                String of tokens separated by "|".
                Each token represents a pattern.
                If wildcard is not None, we return only those files that
                match the given shell pattern (uses fnmatch).

                Example:
                  wildcard="*.nc|*.pdf" selects only those files that end with .nc or .pdf
        """
        # Selecte the files in the directory.
        fnames = [f for f in os.listdir(self.path)]
        filepaths = filter(os.path.isfile, [os.path.join(self.path, f) for f in fnames])

        # Filter using the shell patterns.
        if wildcard is not None:
            filepaths = WildCard(wildcard).filter(filepaths)

        return filepaths

    def has_abiext(self, ext):
        """
        Returns the absolute path of the ABINIT file with extension ext.
        Support both Fortran files and netcdf files. In the later case,
        we check whether a file with extension ext + ".nc" is present 
        in the directory. Returns empty string is file is not present.

        Raises:
            ValueError if multiple files with the given ext are found.
            This implies that this method is not compatible with multiple datasets.
        """
        files = []
        for f in self.list_filepaths():
            if f.endswith(ext) or f.endswith(ext + ".nc"):
                files.append(f)

        if not files:
            return ""

        if len(files) > 1:
            # ABINIT users must learn that multiple datasets are bad!
            err_msg = "Found multiple files with the same extensions\n Please avoid the use of mutiple datasets!"
            raise ValueError(err_msg)

        return files[0]

# This dictionary maps ABINIT file extensions to the 
# variables that must be used to read the file in input.
#
# TODO: It would be nice to pass absolute paths to abinit with getden_path
# so that I can avoid creating symbolic links before running but
# the presence of the C-bindings complicates the implementation
# (gfortran SIGFAULTs if I add strings to dataset_type!
_EXT2VARS = {
    "DEN": {"irdden": 1},
    "WFK": {"irdwfk": 1},
    "SCR": {"irdscr": 1},
    "QPS": {"irdqps": 1},
    "1WF": {"ird1wf": 1},
    "1DEN": {"ird1den": 1},
    "BSR": {"irdbsreso": 1},
    "BSC": {"irdbscoup": 1},
    "HAYDR_SAVE": {"irdhaydock": 1},
    "DDK": {"irdddk": 1},
    "DDB": {},
    "GKK": {},
    "DKK": {},
}

def irdvars_for_ext(ext):
    """
    Returns a dictionary with the ABINIT variables 
    that must be used to read the file with extension ext.
    """
    return _EXT2VARS[ext].copy()


def abi_extensions():
    """List with all the ABINIT extensions that are registered."""
    return list(_EXT2VARS.keys())[:]


def abi_splitext(filename):
    """
    Split the ABINIT extension from a filename.
    "Extension" are found by searching in an internal database.

    Returns "(root, ext)" where ext is the registered ABINIT extension 
    The final ".nc" is included (if any) 

    >>> abi_splitext("foo_WFK")
    ('foo_', 'WFK')

    >>> abi_splitext("/home/guido/foo_bar_WFK.nc")
    ('foo_bar_', 'WFK.nc')
    """
    filename = os.path.basename(filename)
    is_ncfile = False
    if filename.endswith(".nc"):
        is_ncfile = True
        filename = filename[:-3]

    known_extensions = abi_extensions()

    # This algorith fails if we have two files 
    # e.g. HAYDR_SAVE, ANOTHER_HAYDR_SAVE
    for i in range(len(filename)-1, -1, -1):
        ext = filename[i:]
        if ext in known_extensions:
            break

    else:
        raise ValueError("Cannot find a registered extension in %s" % filename)

    root = filename[:i]
    if is_ncfile: 
        ext = ext + ".nc"

    return root, ext


class FilepathFixer(object):
    """
    This object modifies the names of particular output files
    produced by ABINIT so that the file extension is preserved.
    Having a one-to-one mapping between file extension and data format
    is indeed fundamental for the correct behaviour of abinitio since:

        - We locate the output file by just inspecting the extension

        - We select the variables that must be added to the input file
          on the basis of the extension specified by the user during 
          the initialization of the `AbinitFlow`.

    Unfortunately, ABINIT developers like to append extra stuff 
    to the initial extension and therefore we have to call 
    `FilepathFixer` to fix the output files produced by the run.

    Example:
    
    >>> fixer = FilepathFixer()

    >>> fixer.fix_paths('/foo/out_1WF17')
    {'/foo/out_1WF17': '/foo/out_1WF'}

    >>> fixer.fix_paths('/foo/out_1WF5.nc')
    {'/foo/out_1WF5.nc': '/foo/out_1WF.nc'}
    """
    def __init__(self):
        # dictionary mapping the *official* file extension to
        # the regular expression used to tokenize the basename of the file
        # To add a new fix it's sufficient to add a new regexp and 
        # a static method _fix_EXTNAME
        self.regs = regs = {}
        import re
        regs["1WF"] = re.compile("(\w+_)1WF(\d+)(.nc)?$")
        regs["1DEN"] = re.compile("(\w+_)1DEN(\d+)(.nc)?$")

    @staticmethod
    def _fix_1WF(match):
        root, pert, ncext = match.groups()
        if ncext is None: ncext = ""
        return root + "1WF" + ncext

    @staticmethod
    def _fix_1DEN(match):
        root, pert, ncext = match.groups()
        if ncext is None: ncext = ""
        return root + "1DEN" + ncext

    def _fix_path(self, path):
        for ext, regex in self.regs.items():
            head, tail = os.path.split(path)

            match = regex.match(tail)
            if match:
                newtail = getattr(self, "_fix_" + ext)(match)
                newpath = os.path.join(head, newtail)
                return newpath, ext

        return None, None

    def fix_paths(self, paths):
        """
        Fix the filenames in the iterable paths

        Returns:
            old2new:
                Mapping old_path --> new_path
        """
        old2new, fixed_exts = {}, []

        for path in list_strings(paths):
            newpath, ext = self._fix_path(path)

            if newpath is not None:
                assert ext not in fixed_exts
                fixed_exts.append(ext)
                old2new[path] = newpath

        return old2new


def _bop_not(obj):
    """Boolean not."""
    return not bool(obj)

def _bop_and(obj1, obj2):
    """Boolean and."""
    return bool(obj1) and bool(obj2)

def _bop_or(obj1, obj2):
    """Boolean or."""
    return bool(obj1) or bool(obj2)

def _bop_divisible(num1, num2):
    """Return True if num1 is divisible by num2."""
    return (num1 % num2) == 0.0

# Mapping string --> operator.
_UNARY_OPS = {
    "$not": _bop_not,
}

_BIN_OPS = {
    "$eq": operator.eq,
    "$ne": operator.ne,
    "$gt": operator.gt,
    "$ge": operator.ge,
    "$lt": operator.lt,
    "$le": operator.le,
    "$divisible": _bop_divisible,
    "$and": _bop_and,
    "$or":  _bop_or,
    }


_ALL_OPS = list(_UNARY_OPS.keys()) + list(_BIN_OPS.keys())


def map2rpn(map, obj):
    """
    Convert a Mongodb-like dictionary to a RPN list of operands and operators.

    Reverse Polish notation (RPN) is a mathematical notation in which every 
    operator follows all of its operands, e.g.

    3 - 4 + 5 -->   3 4 - 5 + 

    >>> d = {2.0: {'$eq': 1.0}}
    >>> map2rpn(d, None)
    [2.0, 1.0, '$eq']
    """
    rpn = []

    for k, v in map.items():

        if k in _ALL_OPS:
            if isinstance(v, collections.Mapping):
                # e.g "$not": {"$gt": "one"}
                # print("in op_vmap",k, v)
                values = map2rpn(v, obj)
                rpn.extend(values)
                rpn.append(k)

            elif isinstance(v, (list, tuple)):
                # e.g "$and": [{"$not": {"one": 1.0}}, {"two": {"$lt": 3}}]}
                # print("in_op_list",k, v)
                for d in v:
                    rpn.extend(map2rpn(d, obj))

                rpn.append(k)

            else: 
                # Examples
                # 1) "$eq"": "attribute_name"
                # 2) "$eq"": 1.0
                try:
                    #print("in_otherv",k, v)
                    rpn.append(getattr(obj, v))
                    rpn.append(k)

                except TypeError:
                    #print("in_otherv, raised",k, v)
                    rpn.extend([v, k])
        else:
            try:
                k = getattr(obj, k)
            except TypeError:
                k = k

            if isinstance(v, collections.Mapping):
                # "one": {"$eq": 1.0}}
                values = map2rpn(v, obj)
                rpn.append(k)
                rpn.extend(values) 
            else:
                #"one": 1.0 
                rpn.extend([k, v, "$eq"])

    return rpn


def evaluate_rpn(rpn):
    """
    Evaluates the RPN form produced my map2rpn.

    Returns:
        bool
    """
    vals_stack = []
                                                                              
    for item in rpn:
                                                                              
        if item in _ALL_OPS:
            # Apply the operator and push to the task.
            v2 = vals_stack.pop()
                                                                              
            if item in _UNARY_OPS:
                res = _UNARY_OPS[item](v2)
                                                                              
            elif item in _BIN_OPS:
                v1 = vals_stack.pop()
                res = _BIN_OPS[item](v1, v2)
            else:
                raise ValueError("%s not in unary_ops or bin_ops" % str(item))
                                                                              
            vals_stack.append(res)
                                                                              
        else:
            # Push the operand
            vals_stack.append(item)
                                                                              
        #print(vals_stack)
                                                                              
    assert len(vals_stack) == 1
    assert isinstance(vals_stack[0], bool)
                                                                              
    return vals_stack[0]


class Condition(object):
    """
    This object receive a dictionary that defines a boolean condition whose syntax is similar
    to the one used in mongodb (albeit not all the operators available in mongodb are supported here).

    Example:

    $gt: {field: {$gt: value} }

    $gt selects those documents where the value of the field is greater than (i.e. >) the specified value.

    $and performs a logical AND operation on an array of two or more expressions (e.g. <expression1>, <expression2>, etc.) 
    and selects the documents that satisfy all the expressions in the array. 

    { $and: [ { <expression1> }, { <expression2> } , ... , { <expressionN> } ] }

    Consider the following example:

    db.inventory.find( { qty: { $gt: 20 } } )
    This query will select all documents in the inventory collection where the qty field value is greater than 20.
    Consider the following example:

    db.inventory.find( { qty: { $gt: 20 } } )
    db.inventory.find({ $and: [ { price: 1.99 }, { qty: { $lt: 20 } }, { sale: true } ] } )
    """
    def __init__(self, cmap):
        self.cmap = cmap

    def __str__(self):
        return str(self.cmap)

    def apply(self, obj):
        try:
            return evaluate_rpn(map2rpn(self.cmap, obj))
        except:
            logger.warning("Condition.apply() raise Exception")
            return False


class Editor(object):
    """
    Wrapper class that calls the editor specified by the user 
    or the one specified in the $EDITOR env variable.
    """
    def __init__(self, editor=None):
        """If editor is None, $EDITOR is used."""
        if editor is None:
            self.editor = os.getenv("EDITOR", "vi")
        else:
            self.editor = str(editor)

    def edit_files(self, fnames, ask_for_exit=True):
        exit_status = 0
        for idx, fname in enumerate(fnames):
            exit_status = self.edit_file(fname)
            if ask_for_exit and idx != len(fnames)-1 and self.user_wants_to_exit():
                break
        return exit_status

    def edit_file(self, fname):
        from subprocess import call
        retcode = call([self.editor, fname])

        if retcode != 0:
            import warnings
            warnings.warn("Error while trying to edit file: %s" % fname)

        return retcode 

    @staticmethod
    def user_wants_to_exit():
        try:
            answer = raw_input("Do you want to continue [Y/n]")

        except EOFError:
            return True

        return answer.lower().strip() in ["n", "no"]



from __future__ import division

"""
Evaluate the defect concentration based on composition, temperature,
and defect energies using "Dilute Solution Model"
Reference: Phys Rev B, 63, 094103, 2001,
"Density of constitutional and thermal point defects in L12 Al3Sc",
C. Woodward, M. Asta, G. Kresse and J. Hafner.
"""

__author__ = 'Bharat Medasani'
__version__ = "0.2"
__maintainer__ = "Bharat Medasani"
__email__ = "mbkumar@gmail.com"
__status__ = "Alpha"
__date__ = "6/4/14"

import math
import copy
import numpy as np

from monty.dev import requires
from monty.fractions import gcd


try:
    from sympy import Symbol, nsolve, Integer, Rational, Matrix, exp, solve, Eq
    sympy_found = True
except ImportError:
    sympy_found = False

# physical consts
k_B=8.6173324e-5                # eV/K

# Check the inputs
def check_input(def_list):
    flag = True
    for defect in def_list:
        if not defect:
            flag = False
            break
    return flag

# TODO: Cleanup the initialization

@requires(sympy_found,
            "comute_defect_density requires Sympy module. Please install it.")
def dilute_solution_model(structure, e0, vac_defs, antisite_defs, T, 
        trial_chem_pot = None, generate='plot'):

    """
    Compute the defect densities for a structure based on the input parameters
    using dilute solution model.
    Args:
        structure:
            pymatgen.core.structure.Structure object representing the
            primitive or unitcell of the crystal.
        e0:
            The total energy of the undefected system.
            This is E0 from VASP calculation.
        vac_defs:
            List of vacancy defect parameters in the dictionary format.
            The keys of the dict associated with each vacancy defect are
            1) site_index, 2) site_specie, 3) site_multiplicity, and
            4) energy. 1-3 can be obtained from
            pymatgen.analysis.defects.point_defects.Vacancy class.
            Site index is expected to start with 1 (fortran index).
        antisite_defs:
            List of antisite defect parameters in the dictionary format.
            The keys of the dict associated with each antisite defect are
            1) site_index, 2) site_specie, 3) site_multiplicity,
            4) substitution_specie, and 5) energy. 1-3 can be obtained
            from pymatgen.analysis.defects.point_defects.Vacancy class.
        T:
            Temperature in Kelvin
        trial_chem_pot:
            Trial chemical potentials to speedup the plot generation
            Format is {el1:mu1,...}
        generate (string): Options are plot or energy
            Chemical potentials are also returned with energy option.
            If energy option is not chosen, plot is generated.
    """

    if not check_input(vac_defs):
        raise ValueError('Vacancy energy is not defined')
    if not check_input(antisite_defs):
        raise ValueError('Antisite energy is not defined')

    formation_energies = {}
    formation_energies['vacancies'] = copy.deepcopy(vac_defs)
    formation_energies['antisites'] = copy.deepcopy(antisite_defs)
    for vac in formation_energies['vacancies']:
        del vac['energy']
    for asite in formation_energies['antisites']:
        del asite['energy']
    # Setup the system
    site_species = [vac_def['site_specie'] for vac_def in vac_defs]
    multiplicity = [vac_def['site_multiplicity'] for vac_def in vac_defs]
    #print multiplicity
    m = len(set(site_species))      # distinct species
    n = len(vac_defs)           # inequivalent sites

    # Reduce the system and associated parameters such that only distinctive
    # atoms are retained
    comm_div = gcd(*tuple(multiplicity))
    multiplicity = [val/comm_div for val in multiplicity]
    e0 = e0/comm_div
    T = Integer(T)

    c0 = np.diag(multiplicity)
    #print 'c0', c0
    mu = [Symbol('mu'+str(i)) for i in range(m)]

    # Generate maps for hashing
    # Generate specie->mu map and use it for site->mu map
    specie_order = []       # Contains hash for site->mu map    Eg: [Al, Ni]
    site_specie_set = set()             # Eg: {Ni, Al}
    for i in range(n):
        site_specie  = site_species[i]
        if site_specie not in site_specie_set:
            site_specie_set.add(site_specie)
            specie_order.append(site_specie)
    site_mu_map = []     # Eg: [mu0,mu0,mu0,mu1] where mu0->Al, and mu1->Ni
    for i in range(n):
        site_specie  = site_species[i]
        j = specie_order.index(site_specie)
        site_mu_map.append(j)
    specie_site_index_map = []      # Eg: [(0,3),(3,4)] for Al & Ni
    for i in range(m):
        low_ind = site_species.index(specie_order[i])
        if i < m-1:
            hgh_ind = site_species.index(specie_order[i+1])
        else:
            hgh_ind = n
        specie_site_index_map.append((low_ind,hgh_ind))

    #print 'specie_site_index_map', specie_site_index_map
    #for el in specie_site_index_map:
    #    print range(*el)
    #print site_species
    #print 'site_mu_map', site_mu_map
    #print specie_order


    """
    dC: delta concentration matrix:
    dC[i,j,k]: Concentration change of atom i, due to presence of atom
    j on lattice site k
    Special case is [i,i,i] which is considered as vacancy
    Few cases: dC[i,i,i] = -1 due to being vacancy special case
                dC[k,k,i] = +1 due to increment in k at i lattice if i
                               lattice type is of different element
                dC[i,k,i] = -1 due to decrement of ith type atom due to
                presence of kth type atom on ith sublattice and kth type
                atom specie is different from ith sublattice atom specie
                dC[i,k,k] = 0 due to no effect on ith type atom
                dC[i,j,k] = 0 if i!=j!=k
    """
    dC = np.zeros((n,n,n), dtype=np.int)
    for i in range(n):
        for j in range(n):
            for k in range(n):
                if i == j:# and site_species[j] != site_species[k]:
                    dC[i,j,k] = 1
        for j in range(n):
            for k in range(n):
                if i == k:
                    #if j == k or site_species[j] != site_species[k]:
                        dC[i,j,k] = -1

    #print dC
    # dE matrix: Flip energies (or raw defect energies)
    els = [vac_def['site_specie'] for vac_def in vac_defs]
    dE = []
    for i in range(n):
        dE.append([])
    for i in range(n):
        for j in range(n):
            dE[i].append(None)

    for j in range(n):
        for i in range(n):
            if i == j:
                dE[i][j] = vac_defs[i]['energy']
            else:
                sub_specie = vac_defs[i]['site_specie']
                site_specie = vac_defs[j]['site_specie']
                if site_specie == sub_specie:
                    dE[i][j] = 0
                else:
                    for as_def in antisite_defs:
                        if as_def['site_index'] == j+1 and \
                                sub_specie == as_def['substitution_specie']:
                            dE[i][j] = as_def['energy']
                            break
    dE = np.array(dE)
    np.where(dE == None, dE, 0)
    #print dE

    # Initialization for concentrations
    # c(i,p) == presence of ith type atom on pth type site
    c = Matrix(n,n,[0]*n**2)
    for i in range(n):
        for p in range(n):
            c[i,p] = Integer(c0[i,p])
            for epi in range(n):
                sum_mu = sum([mu[site_mu_map[j]]*Integer(
                        dC[j,epi,p]) for j in range(n)])
                c[i,p] += Integer(multiplicity[p]*dC[i,epi,p]) * \
                        exp(-(dE[epi,p]-sum_mu)/(k_B*T))

    #specie_concen = [sum(mult[ind[0]:ind[1]]) for ind in specie_site_index_map]
    #total_c = [sum(c[ind[0]:ind[1]]) for ind in specie_site_index_map]
    total_c = []
    for ind in specie_site_index_map:
        total_c.append(sum([sum(c[i,:]) for i in range(*ind)]))
    #total_c = [sum(c[i,:]) for i in range(n)]
    c_ratio = [total_c[-1]/total_c[i] for i in range(m)]
    #print 'c_ratio'
    #for i in range(len(c_ratio)):
        #print c_ratio[i]

    # Expression for Omega, the Grand Potential
    omega = e0 - sum([mu[site_mu_map[i]]*sum(c0[i,:]) for i in range(n)])
    for p_r in range(n):
        for epi in range(n):
            sum_mu = sum([mu[site_mu_map[j]]*Integer(
                    dC[j,epi,p_r]) for j in range(n)])
            omega -= k_B*T*multiplicity[p_r]*exp(-(dE[epi,p_r]-sum_mu)/(k_B*T))

    def compute_mus():

        def reduce_mu():
            omega = [e0 - sum([mu[site_mu_map[i]]*sum(c0[i,:]) for i in range(n)])]
            x = solve(omega)
            return x

        # Compute trial mu
        mu_red = reduce_mu()
        #print mu_red

        mult = multiplicity
        #for ind in specie_site_index_map:
        #    print ind[0], ind[1]
        specie_concen = [sum(mult[ind[0]:ind[1]]) for ind in specie_site_index_map]
        #print 'specie_concent', specie_concen
        y_vect = [specie_concen[-1]/specie_concen[i] for i in range(m)]
        #print 'y_vect', y_vect
        vector_func = [y_vect[i]-c_ratio[i] for i in range(m-1)]
        vector_func.append(omega)
        #print vector_func
        #vector_func.append(mu_equalities)
        #print 'y0', y0
        min_diff = 1e10
        mu_vals = None
        c_val = None
        m1_min = -20.0
        if e0 > 0:
            m1_max = 10            # Search space needs to be modified
        else:
            m1_max = 0
        for m1 in np.arange(m1_min,m1_max,0.1):
            m0 = mu_red[mu[0]].subs(mu[-1],m1)

            try:
                x = nsolve(vector_func,mu,[m0,m1],module="numpy")
                # Line needs to be modified to include all mus when n > 2
            except:
                continue

            c_val = c.subs(dict(zip(mu,x)))
            #print c_val
            #if all(x >= 0 for x in c_val):
            specie_concen = []
            for ind in specie_site_index_map:
                specie_concen.append(sum([sum(c_val[i,:]) for i in range(*ind)]))
            y_comp = [specie_concen[-1]/specie_concen[i] for i in range(m)]
            diff = math.sqrt(sum([pow(abs(y_comp[i]-y_vect[i]),2) for i in range(m)]))
            if diff < min_diff:
                min_diff = diff
                mu_vals = x

        if mu_vals:
            mu_vals = [float(mu_val) for mu_val in mu_vals]
        else:
            raise ValueError()
        print mu_vals
        return mu_vals
        #print els

    def compute_def_formation_energies():
        i = 0
        for vac_def in vac_defs:
            site_specie = vac_def['site_specie']
            ind = specie_order.index(site_specie)
            uncor_energy = vac_def['energy']
            formation_energy = uncor_energy + mu_vals[ind]
            print site_specie, 'vancancy formation_energy', formation_energy
            formation_energies['vacancies'][i]['formation_energy'] = formation_energy
            specie_ind = site_mu_map[i]
            indices = specie_site_index_map[specie_ind]
            specie_ind_del = indices[1]-indices[0]
            cur_ind = i - indices[0] + 1
            if not specie_ind_del-1:
                label = '$V_{'+site_specie+'}$'
            else:
                label = '$V_{'+site_specie+'_'+str(cur_ind)+'}$'
            formation_energies['vacancies'][i]['label'] = label
            i += 1
        i = 0
        for as_def in antisite_defs:
            site_specie = as_def['site_specie']
            sub_specie = as_def['substitution_specie']
            ind1 = specie_order.index(site_specie)
            ind2 = specie_order.index(sub_specie)
            uncor_energy = as_def['energy']
            formation_energy = uncor_energy + mu_vals[ind1] - mu_vals[ind2]
            print site_specie, sub_specie, 'antisite ', formation_energy
            formation_energies['antisites'][i]['formation_energy'] = formation_energy
            specie_ind = site_mu_map[i]
            indices = specie_site_index_map[specie_ind]
            specie_ind_del = indices[1]-indices[0]
            cur_ind = i - indices[0] + 1
            if not specie_ind_del-1:
                label = '$'+sub_specie+'_{'+site_specie+'}$'
            else:
                label = '$'+sub_specie+'_{'+site_specie+'_'+str(cur_ind)+'}$'
            formation_energies['antisites'][i]['label'] = label
            i += 1
        return formation_energies

    if not trial_chem_pot:
        mu_vals = compute_mus()
    else:
        try:
            mu_vals = [trail_chem_pot[element] for element in specie_ordger]
        except:
            mu_vals = compute_mus()

    if generate == 'energy':
        formation_energies = compute_def_formation_energies()
        mu_dict = dict(zip(specie_order,mu_vals)) 
        return formation_energies, mu_dict


    # Compute ymax
    li = specie_site_index_map[0][0]
    hi = specie_site_index_map[0][1]
    comp1_min = int(sum(multiplicity[li:hi])/sum(multiplicity)*100)-1
    comp2_max = 100-comp1_min
    ymax = comp2_max/comp1_min
    comp1_max = int(sum(multiplicity[li:hi])/sum(multiplicity)*100)+1
    comp2_min = 100-comp1_max
    ymin = comp2_min/comp1_max
    #print ymin, ymax
    delta = (ymax-ymin)/40.0

    #for i in range(len(mu)):
    #    print mu[i], mu_vals[i]

    # Compile mu's for all composition ratios in the range 
    #+/- 1% from the stoichiometry
    result = {}
    for y in np.arange(ymin,ymax,delta):
        result[y] = []
        vector_func = [y-c_ratio[0]]
        vector_func.append(omega)
        x = nsolve(vector_func,mu,mu_vals,module="numpy")
        result[y].append(x[0])
        result[y].append(x[1])


    res = []

    # Compute the concentrations for all the compositions
    for key in result:
        mu_val = result[key]
        total_c_val = [total_c[i].subs(dict(zip(mu,mu_val))) \
                for i in range(len(total_c))]
        c_val = c.subs(dict(zip(mu,mu_val)))
        res1 = []
        # Concentration of first element/over total concen
        res1.append(float(total_c_val[0]/sum(total_c_val)))    
        sum_c0 = sum([c0[i,i] for i in range(n)])
        for i in range(n):
            for j in range(n):
                if i == j:              # Vacancy
                    res1.append(float((c0[i,i]-sum(c_val[:,i]))/c0[i,i]))     
                else:                   # Antisite
                    res1.append(float(c_val[i,j]/c0[i,i]))                    
        res.append(res1)

    res = np.array(res)
    dtype = [('x',np.float64)]+[('y'+str(i)+str(j),np.float64) \
            for i in range(n) for j in range(n)]
    res1 = np.sort(res.view(dtype),order=['x'],axis=0)


    plot_data = {}
    """Because all the plots have identical x-points storing it in a 
    single array"""
    plot_data['x'] = [dat[0][0] for dat in res1]         # x-axis data
    # Element whose composition is varied. For x-label
    plot_data['x_label'] = els[0]+ " mole fraction" 
    plot_data['y_label'] = "Point defect concentration"
    conc = []
    for i in range(n):
        conc.append([])
        for j in range(n):
            conc[i].append([])
    #print conc
    for i in range(n): # Append vacancies
        for j in range(n):
            y1 = [dat[0][i*n+j+1] for dat in res1]
            conc[i][j] = y1
    #print type(conc[i][j])

    y_data = []
    for i in range(n):      # Vacancy plots
        data = conc[i][i]
        specie = els[i]
        specie_ind = site_mu_map[i]
        indices = specie_site_index_map[specie_ind]
        specie_ind_del = indices[1]-indices[0]
        cur_ind = i - indices[0] + 1
        if 'V' not in els:
            vac_string = "$V_{"
        else:
            vac_string = "$Vac_{"
        if not specie_ind_del-1:
            label = vac_string+specie+'}$'
        else:
            label = vac_string+specie+'_'+str(cur_ind)+'}$'
        # Plot data and legend info
        y_data.append({'data':data,'name':label})       

        site_specie = els[i]
        for j in range(m):          # Antisite plot dat
            sub_specie = specie_order[j]
            if sub_specie == site_specie:
                continue
            if not specie_ind_del-1:
                label = '$'+sub_specie+'_{'+specie+'}$'
            else:
                label = '$'+sub_specie+'_{'+specie+'_'+str(cur_ind)+'}$'
            inds = specie_site_index_map[j]
            data = np.sum([conc[ind][i] for ind in range(*inds)],axis=0)
            data = data.tolist()
            y_data.append({'data':data,'name':label})

    plot_data['y'] = y_data

    return plot_data


def compute_defect_density(structure, e0, vac_defs, antisite_defs, T=800, 
        trial_chem_pot=None, plot_style="HighCharts"):
    """
    Wrapper for the dilute_solution_model where the computed plot data is 
    prepared based on plot_style. Only "HighCharts" is supported at this point

    :param structure:
    :param e0:
    :param vac_defs:
    :param antisite_defs:
    :param T:
    :param plot_style:
    :return:
    """
    plot_data = dilute_solution_model(structure,e0,vac_defs,antisite_defs,T,
            trial_chem_pot=trial_chem_pot)

    if plot_style == 'HighCharts':
        hgh_chrt_data = {}
        hgh_chrt_data['xAxis'] = plot_data['x_label']
        hgh_chrt_data['yAxis'] = plot_data['y_label']

        series = []
        x = plot_data['x']
        for y_data in plot_data['y']:
            y = y_data['data']
            xy = zip(x,y)
            xy = [list(el) for el in xy]
            name = y_data['name'].strip('$')
            flds= name.split('_')
            def_string = flds[0]
            site_string = flds[1].strip('{}')
            name = def_string+"<sub>"+site_string+"</sub>"
            #series.append({'data':xy, 'name':y_data['name']})
            series.append({'data':xy, 'name':name})
        hgh_chrt_data['series'] = series
        return hgh_chrt_data



"""
This module contains useful decorators for a variety of functions.
"""
from __future__ import division

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__date__ = "Dec 31, 2011"

import logging

from functools import wraps


def enable_logging(main):
    """
    This decorator is used to decorate main functions.
    It adds the initialization of the logger and an argument parser that allows one to select the loglevel.
    Useful if we are writing simple main functions that call libraries where the logging module is used

    Args:
        main:
            main function.
    """
    @wraps(main)
    def wrapper(*args, **kwargs):
        import argparse
        parser = argparse.ArgumentParser()

        parser.add_argument('--loglevel', default="ERROR", type=str,
                            help="set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")

        options = parser.parse_args()

        # loglevel is bound to the string value obtained from the command line argument.
        # Convert to upper case to allow the user to specify --loglevel=DEBUG or --loglevel=debug
        numeric_level = getattr(logging, options.loglevel.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % options.loglevel)
        logging.basicConfig(level=numeric_level)

        retcode = main(*args, **kwargs)
        return retcode

    return wrapper






"""
This module defines filters for Transmuter object.
"""

from __future__ import division

__author__ = "Will Richards, Shyue Ping Ong, Stephen Dacek"
__copyright__ = "Copyright 2011, The Materials Project"
__version__ = "1.0"
__maintainer__ = "Will Richards"
__email__ = "wrichards@mit.edu"
__date__ = "Sep 25, 2012"

from pymatgen.core.periodic_table import get_el_sp
from pymatgen.serializers.json_coders import MSONable
from pymatgen.analysis.structure_matcher import StructureMatcher,\
    ElementComparator
from pymatgen.symmetry.finder import SymmetryFinder
import abc


class AbstractStructureFilter(MSONable):
    """
    AbstractStructureFilter that defines an API to perform testing of
    Structures. Structures that return True to a test are retained during
    transmutation while those that return False are removed.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def test(self, structure):
        """
        Method to execute the test.

        Returns:
            (bool) Structures that return true are kept in the Transmuter
            object during filtering.
        """
        return

    @classmethod
    def from_dict(cls, d):
        for trans_modules in ['filters']:
            mod = __import__('pymatgen.alchemy.' + trans_modules,
                             globals(), locals(), [d['@class']], -1)
            if hasattr(mod, d['@class']):
                trans = getattr(mod, d['@class'])
                return trans(**d['init_args'])
        raise ValueError("Invalid filter dict")


class ContainsSpecieFilter(AbstractStructureFilter):

    def __init__(self, species, strict_compare=False, AND=True, exclude=False):
        """
        Filter for structures containing certain elements or species.
        By default compares by atomic number

        Args:
            species ([Specie/Element]): list of species to look for
            AND: whether all species must be present to pass (or fail) filter.
            strict_compare: if true, compares objects by specie or element
                object if false, compares atomic number
            exclude: If true, returns false for any structures with the specie
                (excludes them from the Transmuter)
        """
        self._species = map(get_el_sp, species)
        self._strict = strict_compare
        self._AND = AND
        self._exclude = exclude

    def test(self, structure):
        #set up lists to compare
        if not self._strict:
            #compare by atomic number
            atomic_number = lambda x: x.Z
            filter_set = set(map(atomic_number, self._species))
            structure_set = set(map(atomic_number,
                                structure.composition.elements))
        else:
            #compare by specie or element object
            filter_set = set(self._species)
            structure_set = set(structure.composition.elements)

        if self._AND and filter_set <= structure_set:
            #return true if we aren't excluding since all are in structure
            return not self._exclude
        elif (not self._AND) and filter_set & structure_set:
            #return true if we aren't excluding since one is in structure
            return not self._exclude
        else:
            #return false if we aren't excluding otherwise
            return self._exclude

    def __repr__(self):
        return "\n".join(["ContainsSpecieFilter with parameters:",
                          "species = {}".format(self._species),
                          "strict_compare = {}".format(self._strict),
                          "AND = {}".format(self._AND),
                          "exclude = {}".format(self._exclude)])

    @property
    def to_dict(self):
        return {"version": __version__, "@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "init_args": {"species": [str(sp) for sp in self._species],
                              "strict_compare": self._strict,
                              "AND": self._AND,
                              "exclude": self._exclude}}


class SpecieProximityFilter(AbstractStructureFilter):
    """
    This filter removes structures that have certain species that are too close
    together.

    Args:
        specie_and_min_dist_dict: A species string to float mapping. For
            example, {"Na+": 1} means that all Na+ ions must be at least 1
            Angstrom away from each other. Multiple species criteria can be
            applied. Note that the testing is done based on the actual object
            . If you have a structure with Element, you must use {"Na":1}
            instead to filter based on Element and not Specie.

    """

    def __init__(self, specie_and_min_dist_dict):
        self.specie_and_min_dist = {get_el_sp(k): v
                                    for k, v
                                    in specie_and_min_dist_dict.items()}

    def test(self, structure):
        all_species = set(self.specie_and_min_dist.keys())
        for site in structure:
            species = site.species_and_occu.keys()
            sp_to_test = set(species).intersection(all_species)
            if sp_to_test:
                max_r = max([self.specie_and_min_dist[sp]
                             for sp in sp_to_test])
                nn = structure.get_neighbors(site, max_r)
                for sp in sp_to_test:
                    for (nnsite, dist) in nn:
                        if sp in nnsite.species_and_occu.keys():
                            if dist < self.specie_and_min_dist[sp]:
                                return False
        return True

    @property
    def to_dict(self):
        return {"version": __version__, "@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "init_args": {"specie_and_min_dist_dict":
                              {str(sp): v
                              for sp, v in self.specie_and_min_dist.items()}}}


class RemoveDuplicatesFilter(AbstractStructureFilter):
    """
    This filter removes exact duplicate structures from the transmuter.
    """

    def __init__(self, structure_matcher=StructureMatcher(
                 comparator=ElementComparator()), symprec=None):
        """
        Remove duplicate structures based on the structure matcher
        and symmetry (if symprec is given).

        Args:
            structure_matcher: Provides a structure matcher to be used for
                structure comparison.
            symprec: The precision in the symmetry finder algorithm if None (
                default value), no symmetry check is performed and only the
                structure matcher is used. A recommended value is 1e-5.
        """
        self._symprec = symprec
        self._structure_list = []
        if isinstance(structure_matcher, dict):
            self._sm = StructureMatcher.from_dict(structure_matcher)
        else:
            self._sm = structure_matcher

    def test(self, structure):
        if not self._structure_list:
            self._structure_list.append(structure)
            return True

        def get_sg(s):
            finder = SymmetryFinder(s, symprec=self._symprec)
            return finder.get_spacegroup_number()

        for s in self._structure_list:
            if self._sm._comparator.get_structure_hash(structure) ==\
                    self._sm._comparator.get_structure_hash(s):
                if self._symprec is None or \
                        get_sg(s) == get_sg(structure):
                    if self._sm.fit(s, structure):
                        return False

        self._structure_list.append(structure)
        return True

    @property
    def to_dict(self):
        return {"version": __version__, "@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "init_args": {"structure_matcher": self._sm.to_dict}}


class RemoveExistingFilter(AbstractStructureFilter):
    """
    This filter removes structures existing in a given list from the transmuter.
    """

    def __init__(self, existing_structures, structure_matcher=StructureMatcher(
                 comparator=ElementComparator()), symprec=None):
        """
        Remove existing structures based on the structure matcher
        and symmetry (if symprec is given).

        Args:
            existing_structures: List of existing structures to compare with
            structure_matcher: Provides a structure matcher to be used for
                structure comparison.
            symprec: The precision in the symmetry finder algorithm if None (
                default value), no symmetry check is performed and only the
                structure matcher is used. A recommended value is 1e-5.
        """
        self._symprec = symprec
        self._structure_list = []
        self._existing_structures = existing_structures
        if isinstance(structure_matcher, dict):
            self._sm = StructureMatcher.from_dict(structure_matcher)
        else:
            self._sm = structure_matcher

    def test(self, structure):

        def get_sg(s):
            finder = SymmetryFinder(s, symprec=self._symprec)
            return finder.get_spacegroup_number()

        for s in self._existing_structures:
            if self._sm._comparator.get_structure_hash(structure) ==\
                    self._sm._comparator.get_structure_hash(s):
                if self._symprec is None or \
                        get_sg(s) == get_sg(structure):
                    if self._sm.fit(s, structure):
                        return False

        self._structure_list.append(structure)
        return True

    @property
    def to_dict(self):
        return {"version": __version__, "@module": self.__class__.__module__,
                "@class": self.__class__.__name__,
                "init_args": {"structure_matcher": self._sm.to_dict}}


class ChargeBalanceFilter(AbstractStructureFilter):
    """
    This filter removes structures that are not charge balanced from the
    transmuter. This only works if the structure is oxidation state
    decorated, as structures with only elemental sites are automatically
    assumed to have net charge of 0.
    """
    def test(self, structure):
        if structure.charge == 0.0:
            return True
        else:
            return False

    @property
    def to_dict(self):
        return {"@module": self.__class__.__module__,
                "@class": self.__class__.__name__}



"""Strategy objects for creating ABINIT calculations."""
from __future__ import division, print_function

import abc
import collections
import copy
import numpy as np

from pprint import pprint, pformat
from pymatgen.util.string_utils import str_aligned, str_delimited, is_string, list_strings
from pymatgen.io.abinitio.abiobjects import SpinMode, Smearing, Electrons
from pymatgen.io.abinitio.pseudos import PseudoTable

__author__ = "Matteo Giantomassi"
__copyright__ = "Copyright 2013, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Matteo Giantomassi"
__email__ = "gmatteo at gmail.com"


def select_pseudos(pseudos, structure, ret_table=True):
    """
    Given a list of pseudos and a pymatgen structure, extract the pseudopotentials 
    for the calculation (useful when we receive an entire periodic table).

    Raises:
        ValueError if no pseudo is found or multiple occurrences are found.
    """
    table = PseudoTable.astable(pseudos)

    pseudos = []
    for symbol in structure.types_of_specie:
        # Get the list of pseudopotentials in table from atom symbol.
        pseudos_for_type = table.pseudos_with_symbol(symbol)
                                                                             
        if not pseudos_for_type:
            raise ValueError("Cannot find pseudo for symbol %s" % symbol)

        if len(pseudos_for_type) > 1:
            raise ValueError("Find multiple pseudos for symbol %s" % symbol)
                                                                             
        pseudos.append(pseudos_for_type[0])

    if ret_table:
        return PseudoTable(pseudos)
    else:
        return pseudos


def order_pseudos(pseudos, structure):
    return select_pseudos(pseudos, structure, ret_table=False)


def num_valence_electrons(pseudos, structure):
    """
    Compute the number of valence electrons from 
    a list of pseudopotentials and the crystalline structure.

    Args:
        pseudos: 
            List of strings, list of of pseudos or `PseudoTable` instance.
        structure:
            Pymatgen structure.

    Raises:
        ValueError if cannot find a pseudo in the input pseudos or if the
        input list contains more than one pseudo for the chemical symbols
        appearing in structure.
    """
    table = PseudoTable.astable(pseudos)

    valence = 0.0
    for site in structure:
        entries = table.pseudos_with_symbol(site.specie.symbol)
        if len(entries) != 1:
            raise ValueError("Found %d entries for symbol %s" % (len(entries), site.specie.symbol))
        valence += entries[0].Z_val

    return valence


class Strategy(object):
    """
    A Strategy object generates the ABINIT input file used for a particular type of calculation
    e.g. ground-state runs, structural relaxations, self-energy calculations ...

    A Strategy can absorb data (e.g. data produced in the previous steps of a workflow) and 
    can use this piece of information to generate/optimize the input variables.
    Strategy objects must provide the method make_input that builds and returns the abinit input file.

    Attributes:
        accuracy:
            Accuracy of the calculation used to define basic parameters of the run.
            such as tolerances, basis set truncation ... 
        pseudos:
            List of pseudopotentials.
    """
    __metaclass__ = abc.ABCMeta

    # Mapping runlevel --> optdriver variable
    _runl2optdriver = {
        "scf"      : 0 , 
        "nscf"     : 0 ,
        "relax"    : 0 ,
        "dfpt"     : 1 ,
        "screening": 3 ,
        "sigma"    : 4 ,
        "bse"      : 99,
    }

    # Name of the (default) tolerance used by the runlevels.
    _runl2tolname = {
        "scf"      : 'tolvrs', 
        "nscf"     : 'tolwfr', 
        "dfpt"     : 'toldfe',   # ?
        "screening": 'toldfe',   # dummy
        "sigma"    : 'toldfe',   # dummy
        "bse"      : 'toldfe',   # ?
    }

    # Tolerances for the different levels of accuracy.
    T = collections.namedtuple('Tolerance', "low normal high")
    _tolerances = {
        "toldfe": T(1.e-7,  1.e-8,  1.e-9), 
        "tolvrs": T(1.e-7,  1.e-8,  1.e-9),
        "tolwfr": T(1.e-15, 1.e-17, 1.e-19),
        "tolrdf": T(0.04,   0.02,   0.01),
        }
    del T

    def __str__(self):
        return "<%s at %s, accuracy = %s>" % (self.__class__.__name__, id(self), self.accuracy)

    def copy(self):
        """Shallow copy of self."""
        return copy.copy(self)

    def deepcopy(self):
        """Deep copy of self."""
        return copy.deepcopy(self)

    @abc.abstractproperty
    def runlevel(self):
        """String defining the Runlevel. See _runl2optdriver."""
                                                            
    @property
    def optdriver(self):
        """The optdriver associated to the calculation."""
        return self._runl2optdriver[self.runlevel]

    def learn(self, **data):
        """Update the data stored in self."""
        if not hasattr(self, "_data"):
            self._data = dict(data)
        else:
            if [k in self._data for k in data].count(True) != 0:
                raise ValueError("Keys %s are already present in data" % str([k for k in data]))
            self._data.update(data)

    @property
    def accuracy(self):
        """Accuracy used by the strategy."""
        try:
            return self._accuracy
        except AttributeError:
            self.set_accuracy("normal")
            return self._accuracy

    def set_accuracy(self, accuracy):
        """Accuracy setter."""
        if hasattr(self, "_accuracy"):
            raise RuntimeError("object already has accuracy %s " % self._accuracy)

        assert accuracy in ["low", "normal", "high",]
        self._accuracy = accuracy

    @property
    def data(self):
        """Data absorbed by the strategy during the workflow."""
        try:
            return self. _data
        except AttributeError:
            return {}

    @property
    def isnc(self):
        """True if norm-conserving calculation."""
        return self.pseudos.allnc
                                              
    @property
    def ispaw(self):
        """True if PAW calculation."""
        return self.pseudos.allpaw

    @property
    def ecut(self):
        """Cutoff energy in Hartree."""
        try:
            # User option.
            return self.extra_abivars["ecut"] 
        except KeyError:
            # Compute ecut from the Pseudo Hints.
            hints = [p.hint_for_accuracy(self.accuracy) for p in self.pseudos]
            return max(hint.ecut for hint in hints)

    @property
    def pawecutdg(self):
        """Cutoff energy in Hartree for the dense grid used in PAW calculations."""
        if not self.ispaw:
            return None
        try:
            return self.extra_abivars["pawecutdg"] # User option.
        except KeyError:
            raise NotImplementedError("")
            #ratio = max(p.suggested_augratio(accuracy) for p in self.pseudos])
            #ratio = augration_high if high else augratio_norm 
            #pawecutdg = ecut * ratio

    @property
    def tolerance(self):
        """Return a dict {varname: varvalue} with the tolerance used for the calculation."""
        # Check user options first.
        for tolname in self._tolerances:
            try:
                return {tolname: self.extra_abivars[tolname]}
            except KeyError:
                pass

        # Use default values depending on the runlevel and the accuracy.
        tolname = self._runl2tolname[self.runlevel]
        return {tolname: getattr(self._tolerances[tolname], self.accuracy)}

    @property
    def need_forces(self):
        """True if forces are required at each SCF step (like the stresses)."""
        return self.runlevel in ["relax",]
                                                                            
    @property
    def need_stress(self):
        """True if the computation of the stress is required."""
        # TODO: here it's easier to check if optcell != 0
        return self.runlevel in ["relax",]

    def add_extra_abivars(self, abivars):
        """Add variables (dict) to extra_abivars."""
        self.extra_abivars.update(abivars)

    def remove_extra_abivars(self, keys):
        for key in keys:
            self.extra_abivars.pop(key)

    @abc.abstractmethod
    def make_input(self, *args, **kwargs):
        """Returns an Input instance."""

    def num_valence_electrons(self):
        return num_valence_electrons(self.pseudos, self.structure)


class ScfStrategy(Strategy):
    """
    Strategy for ground-state SCF calculations.
    """
    def __init__(self, structure, pseudos, ksampling, accuracy="normal", spin_mode="polarized", 
                 smearing="fermi_dirac:0.1 eV", charge=0.0, scf_algorithm=None, use_symmetries=True, **extra_abivars):
        """
        Args:
            structure:
                pymatgen structure
            pseudos:
                List of pseudopotentials.
            ksampling:
                Ksampling object defining the sampling of the BZ.
            accuracy:
                Accuracy of the calculation.
            spin_mode: 
                Spin polarization mode.
            smearing: 
                string or Smearing instance. 
            charge:
                Total charge of the system. Default is 0.
            scf_algorithm:
                ElectronsAlgorithm instance.
            use_symmetries:
                False if point group symmetries should not be used.
            extra_abivars:
                Extra variables that will be directly added to the input file.
        """
        super(ScfStrategy, self).__init__()

        self.set_accuracy(accuracy)

        self.structure = structure
        self.pseudos   = select_pseudos(pseudos, structure)
        self.ksampling = ksampling
        self.use_symmetries = use_symmetries

        self.electrons  = Electrons(spin_mode=spin_mode,
                                    smearing=smearing,
                                    algorithm=scf_algorithm,
                                    nband=None,
                                    fband=None,
                                    charge=charge,
                                   )

        self.extra_abivars = extra_abivars

    @property
    def runlevel(self):
        return "scf"

    def make_input(self):
        extra = dict(
            optdriver=self.optdriver,
            ecut=self.ecut,
            pawecutdg=self.pawecutdg,
        )
        extra.update(self.tolerance)
        extra.update({"nsym": 1 if not self.use_symmetries else None})

        extra.update(self.extra_abivars)

        input = InputWriter(self.structure, self.electrons, self.ksampling, **extra)
        return input.get_string()


class NscfStrategy(Strategy):
    """
    Strategy for non-self-consistent calculations.
    """
    def __init__(self, scf_strategy, ksampling, nscf_nband, nscf_algorithm=None, **extra_abivars):
        """
        Args:
            scf_strategy:
                ScfStrategy used for the GS run.
            ksampling:
                Ksampling object defining the sampling of the BZ.
            nscf_nband:
                Number of bands to compute.
            nscf_algorithm
                ElectronsAlgorithm instance.
            extra_abivars:
                Extra ABINIT variables that will be directly added to the input file
        """
        super(NscfStrategy, self).__init__()

        self.set_accuracy(scf_strategy.accuracy)

        self.scf_strategy = scf_strategy

        self.nscf_nband = nscf_nband
        self.pseudos    = scf_strategy.pseudos
        self.ksampling  = ksampling

        if nscf_algorithm is None:
            nscf_algorithm = {"iscf": -2}

        # Electrons used in the GS run.
        scf_electrons = scf_strategy.electrons

        self.electrons = Electrons(spin_mode=scf_electrons.spin_mode,
                                   smearing=scf_electrons.smearing,
                                   algorithm=nscf_algorithm,
                                   nband=nscf_nband,
                                   fband=None,
                                   charge=scf_electrons.charge,
                                   comment=None,
                                   #occupancies = None,
                                   )

        self.extra_abivars = extra_abivars

    @property
    def runlevel(self):
        return "nscf"

    def make_input(self):
        # Initialize the system section from structure.
        scf_strategy = self.scf_strategy

        extra = dict(
            optdriver=self.optdriver,
            ecut=self.ecut,
            pawecutdg=self.pawecutdg,
        )
        extra.update(self.tolerance)
        extra.update(self.extra_abivars)
                                                                                     
        input = InputWriter(scf_strategy.structure, self.electrons, self.ksampling, **extra)
        return input.get_string()


class RelaxStrategy(ScfStrategy):
    """Extends ScfStrategy by adding an algorithm for the structural relaxation."""

    def __init__(self, structure, pseudos, ksampling, relax_algo, accuracy="normal", spin_mode="polarized", 
                 smearing="fermi_dirac:0.1 eV", charge=0.0, scf_solver=None, **extra_abivars):
        """
        Args:
            structure:
                pymatgen structure
            pseudos:
                List of pseudopotentials.
            ksampling:
                `Ksampling` object defining the sampling of the BZ.
            relax_algo:
                Object defining the algorithm for the structural relaxation.
            accuracy:
                Accuracy of the calculation.
            spin_mode: 
                Flag defining the spin polarization. Defaults to "polarized"
            smearing: 
                String or `Smearing` instance. 
            charge:
                Total charge of the system. Default is 0.
            scf_algorithm:
                `ElectronsAlgorithm` instance.
            extra_abivars:
                Extra ABINIT variables that will be directly added to the input file
        """
        super(RelaxStrategy, self).__init__(structure, pseudos, ksampling, 
                 accuracy=accuracy, spin_mode=spin_mode, smearing=smearing, 
                 charge=charge, scf_algorithm=scf_algorithm, **extra_abivars)

        self.relax_algo = relax_algo

    @property
    def runlevel(self):
        return "scf"

    def make_input(self):
        # Input for the GS run
        input_str = super(RelaxStrategy, self).make_input()

        # Add the variables for the structural relaxation.
        input = InputWriter(self.relax_algo)
        input_str += input.get_string()

        return input_str


class ScreeningStrategy(Strategy):
    """Strategy for Screening calculations."""

    def __init__(self, scf_strategy, nscf_strategy, screening, **extra_abivars):
        """
        Args:
            scf_strategy:
                Strategy used for the ground-state calculation
            nscf_strategy:
                Strategy used for the non-self consistent calculation
            screening:
                Screening instance
            extra_abivars:
                Extra ABINIT variables added directly to the input file
        """
        super(ScreeningStrategy, self).__init__()

        self.pseudos = scf_strategy.pseudos

        self.scf_strategy = scf_strategy
        self.nscf_strategy = nscf_strategy

        self.screening = screening

        scr_nband = screening.nband

        scf_electrons  = scf_strategy.electrons
        nscf_electrons = nscf_strategy.electrons

        if scr_nband > nscf_electrons.nband:
            raise ValueError("Cannot use more that %d bands for the screening" % nscf_electrons.nband)

        self.ksampling = nscf_strategy.ksampling

        if not self.ksampling.is_homogeneous:
            raise ValueError("The k-sampling used for the NSCF run mush be homogeneous")

        self.electrons = Electrons(spin_mode=scf_electrons.spin_mode,
                                   smearing =scf_electrons.smearing,
                                   nband=scr_nband,
                                   charge=scf_electrons.charge,
                                   comment=None,
                                  )

        self.extra_abivars = extra_abivars

    @property
    def runlevel(self):
        return "screening"

    def make_input(self):
        # FIXME
        extra = dict(
            optdriver=self.optdriver,
            ecut=self.ecut,
            ecutwfn=self.ecut,
            #pawecutdg=self.pawecutdg,
        )
        extra.update(self.tolerance)
        extra.update(self.extra_abivars)
                                                                                     
        input = InputWriter(self.scf_strategy.structure, self.electrons, self.ksampling, self.screening, **extra)
        return input.get_string()


class SelfEnergyStrategy(Strategy):
    """Strategy for self-energy calculations."""
    def __init__(self, scf_strategy, nscf_strategy, scr_strategy, sigma, **extra_abivars):
        """
        Args:
            scf_strategy:
                Strategy used for the ground-state calculation
            nscf_strategy:
                Strategy used for the non-self consistent calculation
            scr_strategy:
                Strategy used for the screening calculation
            sigma:
                SelfEnergy instance.
            extra_abivars:
                Extra ABINIT variables added directly to the input file
        """
        # TODO Add consistency check between SCR and SIGMA strategies
        super(SelfEnergyStrategy, self).__init__()

        self.pseudos = scf_strategy.pseudos

        self.scf_strategy = scf_strategy
        self.nscf_strategy = nscf_strategy
        self.scr_strategy = scr_strategy

        self.sigma = sigma

        self.extra_abivars = extra_abivars

        scf_electrons  = scf_strategy.electrons
        nscf_electrons = nscf_strategy.electrons

        if sigma.nband > nscf_electrons.nband:
            raise ValueError("Cannot use more that %d bands for the self-energy" % nscf_electrons.nband)

        self.ksampling = nscf_strategy.ksampling

        if not self.ksampling.is_homogeneous:
            raise ValueError("The k-sampling used for the NSCF run mush be homogeneous")

        self.electrons = Electrons(spin_mode = scf_electrons.spin_mode,
                                   smearing  = scf_electrons.smearing,
                                   nband     = sigma.nband,
                                   charge    = scf_electrons.charge,
                                   comment   = None,
                                  )
    @property
    def runlevel(self):
        return "sigma"

    def make_input(self):
        # FIXME
        extra = dict(
            optdriver=self.optdriver,
            ecut=self.ecut,
            ecutwfn=self.ecut,
            # "pawecutdg": self.pawecutdg,
            )
        extra.update(self.tolerance)
        extra.update(self.extra_abivars)
                                                                                     
        input = InputWriter(self.scf_strategy.structure, self.electrons, self.ksampling, self.sigma, **extra)
        return input.get_string()


class MDFBSE_Strategy(Strategy):
    """
    Strategy for Bethe-Salpeter calculation based on the 
    model dielectric function and the scissors operator
    """
    def __init__(self, scf_strategy, nscf_strategy, exc_ham, **extra_abivars):
        """
        Args:
            scf_strategy:
                Strategy used for the ground-state calculation.
            nscf_strategy:
                Strategy used for the non-self consistent calculation.
            exc_ham:
                `ExcitonicHamiltonian` instance.
            extra_abivars:
                Extra ABINIT variables added directly to the input file.
        """
        super(MDFBSE_Strategy, self).__init__()

        self.pseudos = scf_strategy.pseudos

        self.scf_strategy = scf_strategy
        self.nscf_strategy = nscf_strategy

        self.exc_ham = exc_ham

        self.extra_abivars = extra_abivars

        scf_electrons  = scf_strategy.electrons
        nscf_electrons = nscf_strategy.electrons

        if exc_ham.nband > nscf_electrons.nband:
            raise ValueError("Cannot use more that %d bands for the EXC hamiltonian." % nscf_electrons.nband)

        self.ksampling = nscf_strategy.ksampling

        if not self.ksampling.is_homogeneous:
            raise ValueError("The k-sampling used for the NSCF run mush be homogeneous")

        self.electrons = Electrons(spin_mode=scf_electrons.spin_mode,
                                   smearing=scf_electrons.smearing,
                                   nband=exc_ham.nband,
                                   charge= scf_electrons.charge,
                                   comment=None,
                                    )
    @property
    def runlevel(self):
        return "bse"

    def make_input(self):
        # FIXME
        extra = dict(
            optdriver=self.optdriver,
            ecut=self.ecut,
            ecutwfn=self.ecut,
            #pawecutdg=self.pawecutdg,
        )
        #extra.update(self.tolerance)
        extra.update(self.extra_abivars)
                                                                                     
        input = InputWriter(self.scf_strategy.structure, self.electrons, self.ksampling, self.exc_ham, **extra)
        return input.get_string()



class InputWriter(object): 
    """
    This object receives a list of `AbivarAble` objects, an optional 
    dictionary with extra ABINIT variables and produces a (nicely formatted?) 
    string with the input file.
    """
    def __init__(self, *args, **kwargs):
        self.abiobj_dict = collections.OrderedDict()
        self.extra_abivars = collections.OrderedDict()

        for arg in args:
            if hasattr(arg, "to_abivars"):
                self.add_abiobj(arg)
            else:
                self.add_extra_abivars(arg)

        for (k,v) in kwargs.items():
            self.add_extra_abivars({k: v})

    def __str__(self):
        """String representation (the section of the abinit input file)."""
        return self.get_string()

    @property
    def abiobjects(self):
        """List of objects stored in self."""
        return self.abiobj_dict.values()

    def add_abiobj(self, obj):
        """Add the object to self."""
        if not hasattr(obj, "to_abivars"):
            raise ValueError("%s does not define the method to_abivars" % str(obj))

        cname = obj.__class__.__name__
        if cname in self.abiobj_dict:
            raise ValueError("%s is already stored" % cname)
        self.abiobj_dict[cname] = obj

    def add_extra_abivars(self, abivars):
        """Add variables (dict) to extra_abivars."""
        self.extra_abivars.update(abivars)

    def to_abivars(self):
        """Returns a dictionary with the abinit variables defined by the Card."""
        abivars = {}
        for obj in self.abiobjects:
            abivars.update(obj.to_abivars())

        abivars.update(self.extra_abivars)
        return abivars

    #def list_objects(self):
    #    "String comment (comment of self + comments of the objects, if any)"
    #    for obj in self.abiobjects:
    #        if hasattr(obj, "comment"):
    #            lines.append("%s: %s" % (obj.__class__.__name__, obj.comment))
    #    return "\n".join(lines)

    @staticmethod
    def _format_kv(key, value):
        """Formatter"""
        if value is None:  
            return [] # Use ABINIT default.
                                                                                   
        if isinstance(value, collections.Iterable) and not is_string(value):
            arr = np.array(value)
            if len(arr.shape) in [0,1]: # scalar or vector.
                token = [key, " ".join(str(i) for i in arr)]
                                                                                   
            else: 
                # array --> matrix 
                matrix = np.reshape(arr, (-1, arr.shape[-1])) 
                lines  = []
                for (idx, row) in enumerate(matrix):
                    lines.append(" ".join(str(i) for i in row))
                token = [key +"\n", "\n".join(lines)]

        else:
            token = [key, str(value)]

        return token

    def get_string(self, pretty=False):
        """
        Returns a string representation of self. The reason why this
        method is different from the __str__ method is to provide options for pretty printing.

        Args:
            pretty:
                Set to True for pretty aligned output.
        """
        lines = []
        app = lines.append

        # Write the Abinit objects first.
        for obj in self.abiobjects:
            app([80*"#", ""])
            app(["#", "%s" % obj.__class__.__name__])
            app([80*"#", ""])
            for (k, v) in obj.to_abivars().items():
                app(self._format_kv(k, v))

        # Extra variables.
        if self.extra_abivars:
            app([80*"#", ""])
            app(["#", "Extra_Abivars"])
            app([80*"#", ""])
            for (k, v) in self.extra_abivars.items():
                app(self._format_kv(k, v))

        if pretty:
            return str_aligned(lines, header=None)
        else:
            return str_delimited(lines, header=None, delimiter=5*" ")


class StrategyWithInput(object):
    # TODO: Find a better way to do this. I will likely need to refactor the Strategy object
    def __init__(self, abinit_input):
        self.abinit_input = abinit_input
                                         
    @property
    def pseudos(self):
        # FIXME: pseudos must be order but I need to define an ABC for the Strategies and Inputs.
        # Order pseudos
        pseudos = self.abinit_input.pseudos
        return order_pseudos(pseudos, self.abinit_input.structure)
        #print("pseudos", pseudos)
        #print("ord_pseudos", ord_pseudos)
        #return ord_pseudos
        #return self.abinit_input.pseudos

    def add_extra_abivars(self, abivars):
        """Add variables (dict) to extra_abivars."""
        self.abinit_input.set_variables(**abivars)

    def remove_extra_abivars(self, keys):
        """Remove variables from extra_abivars."""
        self.abinit_input.remove_variables(keys)
                                         
    def make_input(self):
        return str(self.abinit_input)


class OpticVar(collections.namedtuple("OpticVar", "name value help")):
    def __str__(self):
        sval = string(self.value)
        return (4*" ").join(sval, "!" + self.help)

class OpticInput(object):
    """
    abo_1WF7      ! Name of the first d/dk response wavefunction file, produced by abinit
    abo_1WF8      ! Name of the second d/dk response wavefunction file, produced by abinit
    abo_1WF9      ! Name of the third d/dk response wavefunction file, produced by abinit
    abo_WFK       ! Name of the ground-state wavefunction file, produced by abinit
    0.01          ! Value of the *smearing factor*, in Hartree
    0.010   1     ! frequency *step* and *maximum* frequency (Ha)
    0.000         ! *Scissor* shift if needed, in Hartree
    0.001         ! *Tolerance* on closeness of singularities (in Hartree)
    3             ! *Number of components* of linear optic tensor to be computed
    11 33 23      ! Linear *coefficients* to be computed (x=1, y=2, z=3)
    2             ! Number of components of nonlinear optic tensor to be computed
    123 222       ! Non-linear coefficients to be computed
    """

    # variable name --> default value.
    _VARIABLES = [
        OpticVar("ddkfile_x", None,  "Name of the first d/dk response wavefunction file"),
        OpticVar("ddkfile_y", None,  "Name of the second d/dk response wavefunction file"),
        OpticVar("ddkfile_z", None,  "Name of the third d/dk response wavefunction file"),
        OpticVar("wfkfile",   None,  "Name of the ground-state wavefunction file"),
        OpticVar("zcut",      0.01,  "Value of the *smearing factor*, in Hartree"),
        OpticVar("wmesh",(0.010, 1), "Frequency *step* and *maximum* frequency (Ha)"),
        OpticVar("scissor", 0.000,   "*Scissor* shift if needed, in Hartree"),
        OpticVar("sing_tol", 0.001,  "*Tolerance* on closeness of singularities (in Hartree)"), 
        OpticVar("num_lin_comp", None, "*Number of components* of linear optic tensor to be computed"),
        OpticVar("lin_comp", None,    "Linear *coefficients* to be computed (x=1, y=2, z=3)"),
        OpticVar("num_nonlin_comp", None, "Number of components of nonlinear optic tensor to be computed"),
        OpticVar("nonlin_comp", None, "! Non-linear coefficients to be computed"),
    ]                                             

    _VARNAMES = [v.name for v in _VARIABLES]

    def __init__(self, **kwargs):
        # Default values
        self.vars = collections.OrderedDict((v.name, v.value) for v in _VARIABLES)

        # Update the variables with the values passed by the user
        for k, v in kwargs:
            if k not in self.VARNAMES:
                raise ValueError("varname %s not in %s" % (k, str(self.VARNAMES)))
            self.vars[k] = v

    def __init__(self, zcut, wstep, wmax, scissor, sing_tol, linear_components, 
                nonlinear_components=None, ddk_files=None, wfk=None): 

        self.vars = vars = collections.OrderedDict(*self.VAR_NAMES)

        if ddk_files is not None:
            assert len(ddk_files) == 3
            assert wfk is not None
            for dir, ddk in zip(["x", "y", "z"], ddk_files):
                vars["ddkfile_" + dir] = os.path.abspath(ddk)

        if wfk is not None:
            vars["wfkfile"] = os.path.abspath(wfk)

        vars["zcut"] = zcut
        vars["wmesh"] = " ".join(map(str, (wstep, wmax)))
        vars["sing_tol"] = sing_tol

        vars["num_lin_comp"] = len(linear_components)
        vars["lin_comp"] = " ".join(str(c) for c in linear_components)

        vars["num_nonlin_comp"] = len(non_linear_components)
        vars["nonlin_comp"] = " ".join(str(c) for c in nonlinear_components)

    def __init__(self, string):
        self.string = string

    def __str__(self):
        return self.string

    def to_string(self):
        lines = []
        app = lines.append

        for name in self.VARNAMES:
            var = self.vars[name]
            app(str(var))

        return "\n".join(lines)

    def make_input(self):
        return str(self)

    def add_extra_abivars(self, abivars):
        """
        Connection is explicit via the input file
        since we can pass the paths of the output files 
        produced by the previous runs.
        """


class AnaddbInput(object):

    def __init__(self, structure=None, ndtset=1, comment=""):
        """
        Args:
            structure:
                Crystalline structure.
            ndtset:
                Number of datasets.
            comment:
                Optional string with a comment that will be placed at the beginning of the file.
        """
        self._structure = _structure
        self.ndtset = ndset
        self.comment = comment

    def __init__(self, string):
        self.string = string

    @property
    def structure(self):
        return self._structure

    def to_string(self):
        return self.string

    def make_input(self):
        return self.to_string()

    def add_extra_abivars(self, abivars):
        """
        Connection is explicit via the input file
        since we can pass the paths of the output files 
        produced by the previous runs.
        """

    #def set_qpath(self):
    #    """
    #    nqpath 7
    #    qpath
    #     0.0 0.0 0.0
    #     1/2 1/2 0.0
    #     1   1   1
    #     1/2 1/2 1/2
    #     1/2 1/2 0.0
    #     1/2 3/4 1/4
    #     1/2 1/2 1/2
    #     """

    #def split_datasets(self):
    #    """
    #    Split an input file with multiple datasets into a  list of `ndtset` distinct input files.
    #    """
    #    # Propagate subclasses (if any)
    #    cls = self.__class__ 
    #    news = []

    #    for i in range(self.ndtset):
    #        my_vars = self[i+1].allvars
    #        my_vars.pop("ndtset", None)

    #        new = cls(pseudos=self.pseudos, ndtset=1)
    #        new.set_variables(**my_vars)
    #        news.append(new)
    #
    #    return news



"""
A module to perform diffusion analyses (e.g. calculating diffusivity from
mean square displacements etc.). If you use this module, please consider
citing the following papers::

    Ong, S. P., Mo, Y., Richards, W. D., Miara, L., Lee, H. S., & Ceder, G.
    (2013). Phase stability, electrochemical stability and ionic conductivity
    of the Li10+-1MP2X12 (M = Ge, Si, Sn, Al or P, and X = O, S or Se) family
    of superionic conductors. Energy & Environmental Science, 6(1), 148.
    doi:10.1039/c2ee23355j

    Mo, Y., Ong, S. P., & Ceder, G. (2012). First Principles Study of the
    Li10GeP2S12 Lithium Super Ionic Conductor Material. Chemistry of Materials,
    24(1), 15-17. doi:10.1021/cm203303y
"""

from __future__ import division

__author__ = "Will Richards"
__version__ = "0.1"
__maintainer__ = "Will Richards"
__email__ = "wrichard@mit.edu"
__status__ = "Beta"
__date__ = "5/2/13"

import math

import numpy as np

from pymatgen.core import Structure, get_el_sp
import pymatgen.core.physical_constants as phyc
from pymatgen.serializers.json_coders import MSONable
from pymatgen.io.vaspio.vasp_output import Vasprun
from pymatgen.util.coord_utils import pbc_diff


class DiffusionAnalyzer(MSONable):
    """
    Class for performing diffusion analysis.

    .. attribute: diffusivity

        Diffusivity in cm^2 / cm

    .. attribute: conductivity

        Conductivity in mS / cm

    .. attribute: diffusivity_components

        A vector with diffusivity in the a, b and c directions in cm^2 / cm

    .. attribute: conductivity components

        A vector with conductivity in the a, b and c directions in mS / cm

    .. attribute: max_framework_displacement

        The maximum (drift adjusted) distance of any framework atom from its
        starting location in A

    """

    def __init__(self, structure, displacements, specie, temperature,
                 time_step, step_skip, min_obs=30, weighted=True):
        """
        This constructor is meant to be used with pre-processed data.
        Other convenient constructors are provided as class methods (see
        from_vaspruns and from_files).

        Given a matrix of displacements (see arguments below for expected
        format), the diffusivity is given by:

        D = 1 / 2dt * <mean square displacement>

        where d is the dimensionality, t is the time. To obtain a reliable
        diffusion estimate, a least squares regression of the MSD against
        time to obtain the slope, which is then related to the diffusivity.

        Args:
            structure (Structure): Initial structure.
            displacements (array): Numpy array of with shape [site,
                time step, axis]
            specie (Element/Specie): Specie to calculate diffusivity for as a
                String. E.g., "Li".
            temperature (float): Temperature of the diffusion run in Kelvin.
            time_step (int): Time step between measurements.
            step_skip (int): Sampling frequency of the displacements (
                time_step is multiplied by this number to get the real time
                between measurements)
            min_obs (int): Minimum number of observations to have before
                including in the MSD vs dt calculation. E.g. If a structure
                has 10 diffusing atoms, and min_obs = 30, the MSD vs dt will be
                calculated up to dt = total_run_time / 3, so that each
                diffusing atom is measured at least 3 uncorrelated times.
            weighted (bool): Uses a weighted least squares to fit the
                MSD vs dt. Weights are proportional to 1/dt, since the
                number of observations are also proportional to 1/dt (and
                hence the variance is proportional to dt)
        """
        self.s = structure
        self.disp = displacements
        self.sp = specie
        self.temperature = temperature
        self.time_step = time_step
        self.step_skip = step_skip
        self.min_obs = min_obs
        self.weighted = weighted

        self.indices = []
        self.framework_indices = []

        for i, site in enumerate(structure):
            if site.specie.symbol == specie:
                self.indices.append(i)
            else:
                self.framework_indices.append(i)
        if self.disp.shape[1] < 2:
            self.diffusivity = 0.
            self.conductivity = 0.
            self.diffusivity_components = np.array([0., 0., 0.])
            self.conductivity_components = np.array([0., 0., 0.])
            self.max_framework_displacement = 0
        else:
            framework_disp = self.disp[self.framework_indices]
            drift = np.average(framework_disp, axis=0)[None, :, :]

            #drift corrected position
            dc_x = self.disp[self.indices] - drift
            dc_framework = self.disp[self.framework_indices] - drift
            self.max_framework_displacement = \
                np.max(np.sum(dc_framework ** 2, axis=-1) ** 0.5)
            df_x = self.s.lattice.get_fractional_coords(dc_x)

            #limit the number of sampled timesteps to 200
            min_dt = int(1000 / (self.step_skip * self.time_step))
            max_dt = min(dc_x.shape[0] * dc_x.shape[1] // self.min_obs,
                         dc_x.shape[1])
            if min_dt >= max_dt:
                raise ValueError('Not enough data to calculate diffusivity')
            timesteps = np.arange(min_dt, max_dt,
                                  max(int((max_dt - min_dt) / 200), 1))
            self.dt = timesteps * self.time_step * self.step_skip

            #calculate the smoothed msd values
            self.s_msd = np.zeros_like(self.dt, dtype=np.double)
            self.s_msd_components = np.zeros(self.dt.shape + (3,))
            lengths = np.array(self.s.lattice.abc)[None, None, :]
            for i, n in enumerate(timesteps):
                dx = dc_x[:, n:, :] - dc_x[:, :-n, :]
                self.s_msd[i] = 3 * np.average(dx ** 2)
                dcomponents = (df_x[:, n:, :] - df_x[:, :-n, :]) * lengths
                self.s_msd_components[i] = np.average(
                    np.average(dcomponents ** 2, axis=1), axis=0)

            #run the regression on the msd components
            if weighted:
                w = 1 / self.dt
            else:
                w = np.ones_like(self.dt)

            #weighted least squares
            def weighted_lstsq(a, b, w):
                w_root = w ** 0.5
                x, res, rank, s = np.linalg.lstsq(a * w_root[:, None],
                                                  b * w_root)
                return x

            m_components = np.zeros(3)
            a = np.ones((len(self.dt), 2))
            a[:, 0] = self.dt
            for i in range(3):
                (m, c) = weighted_lstsq(a, self.s_msd_components[:, i], w)
                m_components[i] = max(m, 1e-15)

            (m, c) = weighted_lstsq(a, self.s_msd, w)
            #m shouldn't be negative
            m = max(m, 1e-15)

            #factor of 10 is to convert from A^2/fs to cm^2/s
            #factor of 6 is for dimensionality
            conv_factor = get_conversion_factor(self.s, self.sp,
                                                self.temperature)
            self.diffusivity = m / 60
            self.conductivity = self.diffusivity * conv_factor

            self.diffusivity_components = m_components / 20
            self.conductivity_components = self.diffusivity_components * \
                conv_factor

    def get_summary_dict(self, include_msd_t=False):
        """
        Provides a summary of diffusion information.

        Args:
            include_msd_t (bool): Whether to include mean square displace and
                time data with the data.

        Returns:
            (dict) of diffusion and conductivity data.
        """
        d = {
            "D": self.diffusivity,
            "S": self.conductivity,
            "D_components": self.diffusivity_components.tolist(),
            "S_components": self.conductivity_components.tolist(),
            "specie": str(self.sp),
            "step_skip": self.step_skip,
            "time_step": self.time_step,
            "temperature": self.temperature
        }
        if include_msd_t:
            d["msd"] = self.s_msd.tolist()
            d["msd_components"] = self.s_msd_components.tolist()
            d["dt"] = self.dt.tolist()
        return d

    def get_smoothed_msd_plot(self, plt=None):
        """
        Get the plot of the smoothed msd vs time graph. Useful for
        checking convergence. This can be written to an image file.

        Args:
            plt: A plot object. Defaults to None, which means one will be
                generated.
        """
        from pymatgen.util.plotting_utils import get_publication_quality_plot
        plt = get_publication_quality_plot(12, 8, plt=plt)
        plt.plot(self.dt, self.s_msd, 'k')
        plt.plot(self.dt, self.s_msd_components[:, 0], 'r')
        plt.plot(self.dt, self.s_msd_components[:, 1], 'g')
        plt.plot(self.dt, self.s_msd_components[:, 2], 'b')
        plt.legend(["Overall", "a", "b", "c"], loc=2, prop={"size": 20})
        plt.xlabel("Timestep")
        plt.ylabel("MSD")
        plt.tight_layout()
        return plt

    def plot_smoothed_msd(self):
        """
        Plot the smoothed msd vs time graph. Useful for checking convergence.
        """
        self.get_smoothed_msd_plot().show()

    @classmethod
    def from_structures(cls, structures, specie, temperature,
                        time_step, step_skip, min_obs=30, weighted=True):
        """
        Convenient constructor that takes in a list of Structure objects to
        perform diffusion analysis.

        Args:
            structures ([Structure]): list of Structure objects (must be
                ordered in sequence of run). E.g., you may have performed
                sequential VASP runs to obtain sufficient statistics.
            specie (Element/Specie): Specie to calculate diffusivity for as a
                String. E.g., "Li".
            temperature (float): Temperature of the diffusion run in Kelvin.
            time_step (int): Time step between measurements.
            step_skip (int): Sampling frequency of the displacements (
                time_step is multiplied by this number to get the real time
                between measurements)
            min_obs (int): Minimum number of observations to have before
                including in the MSD vs dt calculation. E.g. If a structure
                has 10 diffusing atoms, and min_obs = 30, the MSD vs dt will be
                calculated up to dt = total_run_time / 3, so that each
                diffusing atom is measured at least 3 uncorrelated times.
            weighted (bool): Uses a weighted least squares to fit the
                MSD vs dt. Weights are proportional to 1/dt, since the
                number of observations are also proportional to 1/dt (and
                hence the variance is proportional to dt)
        """
        structure = structures[0]

        p = [np.array(s.frac_coords)[:, None] for s in structures]
        p = np.concatenate(p, axis=1)
        dp = p[:, 1:] - p[:, :-1]
        dp = np.concatenate([np.zeros_like(dp[:, (0,)]), dp], axis=1)
        dp = dp - np.round(dp)
        f_disp = np.cumsum(dp, axis=1)
        disp = structure.lattice.get_cartesian_coords(f_disp)

        return cls(structure, disp, specie, temperature,
                   time_step, step_skip=step_skip, min_obs=min_obs,
                   weighted=weighted)

    @classmethod
    def from_vaspruns(cls, vaspruns, specie, min_obs=30, weighted=True):
        """
        Convenient constructor that takes in a list of Vasprun objects to
        perform diffusion analysis.

        Args:
            vaspruns ([Vasprun]): List of Vaspruns (must be ordered  in
                sequence of MD simulation). E.g., you may have performed
                sequential VASP runs to obtain sufficient statistics.
            specie (Element/Specie): Specie to calculate diffusivity for as a
                String. E.g., "Li".
            min_obs (int): Minimum number of observations to have before
                including in the MSD vs dt calculation. E.g. If a structure
                has 10 diffusing atoms, and min_obs = 30, the MSD vs dt will be
                calculated up to dt = total_run_time / 3, so that each
                diffusing atom is measured at least 3 uncorrelated times.
            weighted (bool): Uses a weighted least squares to fit the
                MSD vs dt. Weights are proportional to 1/dt, since the
                number of observations are also proportional to 1/dt (and
                hence the variance is proportional to dt)
        """
        step_skip = vaspruns[0].ionic_step_skip or 1

        final_structure = vaspruns[0].initial_structure
        structures = []
        for vr in vaspruns:
            #check that the runs are continuous
            fdist = pbc_diff(vr.initial_structure.frac_coords,
                             final_structure.frac_coords)
            if np.any(fdist > 0.001):
                raise ValueError('initial and final structures do not '
                                 'match.')
            final_structure = vr.final_structure

            assert (vr.ionic_step_skip or 1) == step_skip
            structures.extend([s['structure'] for s in vr.ionic_steps])

        temperature = vaspruns[0].parameters['TEEND']
        time_step = vaspruns[0].parameters['POTIM']

        return cls.from_structures(structures, specie, temperature,
                   time_step, step_skip=step_skip, min_obs=min_obs,
                   weighted=weighted)

    @classmethod
    def from_files(cls, filepaths, specie, step_skip=10, min_obs=30,
                   weighted=True, ncores=None):
        """
        Convenient constructor that takes in a list of vasprun.xml paths to
        perform diffusion analysis.

        Args:
            filepaths ([str]): List of paths to vasprun.xml files of runs. (
                must be ordered in sequence of MD simulation). For example,
                you may have done sequential VASP runs and they are in run1,
                run2, run3, etc. You should then pass in
                ["run1/vasprun.xml", "run2/vasprun.xml", ...].
            specie (Element/Specie): Specie to calculate diffusivity for as a
                String. E.g., "Li".
            step_skip (int): Sampling frequency of the displacements (
                time_step is multiplied by this number to get the real time
                between measurements)
            min_obs (int): Minimum number of observations to have before
                including in the MSD vs dt calculation. E.g. If a structure
                has 10 diffusing atoms, and min_obs = 30, the MSD vs dt will be
                calculated up to dt = total_run_time / 3, so that each
                diffusing atom is measured at least 3 uncorrelated times.
            weighted (bool): Uses a weighted least squares to fit the
                MSD vs dt. Weights are proportional to 1/dt, since the
                number of observations are also proportional to 1/dt (and
                hence the variance is proportional to dt)
            ncores (int): Numbers of cores to use for multiprocessing. Can
                speed up vasprun parsing considerably. Defaults to None,
                which means serial. It should be noted that if you want to
                use multiprocessing, the number of ionic steps in all vasprun
                .xml files should be a multiple of the ionic_step_skip.
                Otherwise, inconsistent results may arise. Serial mode has no
                such restrictions.
        """
        if ncores is not None:
            import multiprocessing
            p = multiprocessing.Pool(ncores)
            vaspruns = p.map(_get_vasprun, [(p, step_skip) for p in filepaths])
        else:
            vaspruns = []
            offset = 0
            for p in filepaths:
                v = Vasprun(p, ionic_step_offset=offset,
                            ionic_step_skip=step_skip)
                vaspruns.append(v)
                # Recompute offset.
                offset = (- (v.nionic_steps - offset)) % step_skip
        return cls.from_vaspruns(vaspruns, min_obs=min_obs,
                                 weighted=weighted, specie=specie)

    @property
    def to_dict(self):
        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "structure": self.s.to_dict,
            "displacements": self.disp.tolist(),
            "specie": self.sp,
            "temperature": self.temperature,
            "time_step": self.time_step,
            "step_skip": self.step_skip,
            "min_obs": self.min_obs,
            "weighted": self.weighted
        }

    @classmethod
    def from_dict(cls, d):
        structure = Structure.from_dict(d["structure"])
        return cls(structure, np.array(d["displacements"]), specie=d["specie"],
                   temperature=d["temperature"], time_step=d["time_step"],
                   step_skip=d["step_skip"], min_obs=d["min_obs"],
                   weighted=d["weighted"])


def get_conversion_factor(structure, species, temperature):
    """
    Conversion factor to convert between cm^2/s diffusivity measurements and
    mS/cm conductivity measurements based on number of atoms of diffusing
    species. Note that the charge is based on the oxidation state of the
    species (where available), or else the number of valence electrons
    (usually a good guess, esp for main group ions).

    Args:
        structure (Structure): Input structure.
        species (Element/Specie): Diffusing species.
        temperature (float): Temperature of the diffusion run in Kelvin.

    Returns:
        Conversion factor.
        Conductivity (in mS/cm) = Conversion Factor * Diffusivity (in cm^2/s)
    """
    df_sp = get_el_sp(species)
    if hasattr(df_sp, "oxi_state"):
        z = df_sp.oxi_state
    else:
        z = df_sp.full_electronic_structure[-1][2]

    n = structure.composition[species]

    V = structure.volume * 1e-24  # units cm^3
    return 1000 * n / (V * phyc.N_a) * z ** 2 * phyc.F ** 2\
        / (phyc.R * temperature)


def _get_vasprun(args):
    """
    Internal method to support multiprocessing.
    """
    return Vasprun(args[0], ionic_step_skip=args[1])


def get_extrapolated_diffusivity(temps, diffusivities, new_temp):
    """
    Returns (Arrhenius) extrapolated diffusivity at new_temp

    Args:
        temps ([float]): A sequence of temperatures. units: K
        diffusivities ([float]): A sequence of diffusivities (e.g.,
            from DiffusionAnalyzer.diffusivity). units: cm^2/s
        new_temp (float): desired temperature. units: K

    Returns:
        (float) Diffusivity at extrapolated temp in mS/cm.
    """
    t_1 = 1000 / np.array(temps)
    logd = np.log10(diffusivities)
    #Do a least squares regression of log(D) vs 1000/T
    A = np.array([t_1, np.ones(len(temps))]).T
    w = np.array(np.linalg.lstsq(A, logd)[0])
    return 10 ** (w[0] * 1000 / new_temp + w[1])


def get_extrapolated_conductivity(temps, diffusivities, new_temp, structure,
                                  species):
    """
    Returns extrapolated mS/cm conductivity.

    Args:
        temps ([float]): A sequence of temperatures. units: K
        diffusivities ([float]): A sequence of diffusivities (e.g.,
            from DiffusionAnalyzer.diffusivity). units: cm^2/s
        new_temp (float): desired temperature. units: K
        structure (structure): structure used for the diffusivity calculation
        species (string/Specie): conducting species

    Returns:
        (float) Conductivity at extrapolated temp in mS/cm.
    """
    return get_extrapolated_diffusivity(temps, diffusivities, new_temp) \
        * get_conversion_factor(structure, species, new_temp)


def get_arrhenius_plot(temps, diffusivities, **kwargs):
    """
    Returns an Arrhenius plot.

    Args:
        temps ([float]): A sequence of temperatures.
        diffusivities ([float]): A sequence of diffusivities (e.g.,
            from DiffusionAnalyzer.diffusivity).
        \*\*kwargs:
            Any keyword args supported by matplotlib.pyplot.plot.

    Returns:
        A matplotlib.pyplot object. Do plt.show() to show the plot.
    """
    t_1 = 1000 / np.array(temps)
    logd = np.log10(diffusivities)
    #Do a least squares regression of log(D) vs 1000/T
    A = np.array([t_1, np.ones(len(temps))]).T
    w = np.array(np.linalg.lstsq(A, logd)[0])
    from pymatgen.util.plotting_utils import get_publication_quality_plot
    plt = get_publication_quality_plot(12, 8)
    plt.plot(t_1, logd, 'ko', t_1, np.dot(A, w), 'k--', markersize=10,
             **kwargs)
    # Calculate the activation energy in meV = negative of the slope,
    # * kB (/ electron charge to convert to eV), * 1000 (inv. temperature
    # scale), * 1000 (eV -> meV), * math.log(10) (the regression is carried
    # out in base 10 for easier reading of the diffusivity scale,
    # but the Arrhenius relationship is in base e).
    actv_energy = - w[0] * phyc.k_b / phyc.e * 1e6 * math.log(10)
    plt.text(0.6, 0.85, "E$_a$ = {:.0f} meV".format(actv_energy),
             fontsize=30, transform=plt.axes().transAxes)
    plt.ylabel("log(D (cm$^2$/s))")
    plt.xlabel("1000/T (K$^{-1}$)")
    plt.tight_layout()
    return plt
